﻿namespace Json2CSharpCodeGenerator.Lib.CodeWriters;

public class TypeScriptCodeWriter : ICodeWriter
{
    public string FileExtension => ".ts";

    public string DisplayName => "TypeScript";

    public string GetTypeName(JsonType type, IJsonClassGeneratorConfig config)
    {
        return type.Type switch
        {
            JsonTypeEnum.Anything => "any",
            JsonTypeEnum.String => "string",
            JsonTypeEnum.Boolean => "bool",
            JsonTypeEnum.Integer or JsonTypeEnum.Long or JsonTypeEnum.Float => "number",
            JsonTypeEnum.Date => "Date",
            JsonTypeEnum.NullableInteger or JsonTypeEnum.NullableLong or JsonTypeEnum.NullableFloat => "number",
            JsonTypeEnum.NullableBoolean => "bool",
            JsonTypeEnum.NullableDate => "Date",
            JsonTypeEnum.Object => type.AssignedName,
            JsonTypeEnum.Array => GetTypeName(type.InternalType, config) + "[]",
            JsonTypeEnum.Dictionary => "{ [key: string]: " + GetTypeName(type.InternalType, config) + "; }",
            JsonTypeEnum.NullableSomething => "any",
            JsonTypeEnum.NonConstrained => "any",
            _ => throw new NotSupportedException("Unsupported type"),
        };
    }

    public void WriteClass(IJsonClassGeneratorConfig config, TextWriter sw, JsonType type)
    {
        string prefix = GetNamespace(config, type.IsRoot) != null ? "    " : "";
        bool exported = !config.InternalVisibility || config.SecondaryNamespace != null;
        sw.WriteLine(prefix + (exported ? "export " : string.Empty) + "interface " + type.AssignedName + " {");
        foreach (FieldInfo field in type.Fields)
        {
            bool shouldDefineNamespace = type.IsRoot && config.SecondaryNamespace != null && config.Namespace != null && (field.Type.Type == JsonTypeEnum.Object || (field.Type.InternalType != null && field.Type.InternalType.Type == JsonTypeEnum.Object));
            if (config.ExamplesInDocumentation)
            {
                sw.WriteLine();
                sw.WriteLine(prefix + "    /**");
                sw.WriteLine(prefix + "      * Examples: " + field.GetExamplesText());
                sw.WriteLine(prefix + "      */");
            }

            sw.WriteLine(prefix + "    " + field.JsonMemberName + (IsNullable(field.Type.Type) ? "?" : "") + ": " + (shouldDefineNamespace ? config.SecondaryNamespace + "." : string.Empty) + GetTypeName(field.Type, config) + ";");
        }
        sw.WriteLine(prefix + "}");
        sw.WriteLine();
    }

    private bool IsNullable(JsonTypeEnum type)
    {
        return
            type is JsonTypeEnum.NullableBoolean or
            JsonTypeEnum.NullableDate or
            JsonTypeEnum.NullableFloat or
            JsonTypeEnum.NullableInteger or
            JsonTypeEnum.NullableLong or
            JsonTypeEnum.NullableSomething;
    }

    public void WriteFileStart(IJsonClassGeneratorConfig config, TextWriter sw)
    {
        // foreach (var line in JsonClassGenerator.FileHeader)
        // {
        //     sw.WriteLine("// " + line);
        // }
        // sw.WriteLine();
    }

    public void WriteFileEnd(IJsonClassGeneratorConfig config, TextWriter sw)
    {
    }

    private string GetNamespace(IJsonClassGeneratorConfig config, bool root) => root ? config.Namespace : (config.SecondaryNamespace ?? config.Namespace);

    public void WriteNamespaceStart(IJsonClassGeneratorConfig config, TextWriter sw, bool root)
    {
        if (GetNamespace(config, root) != null)
        {

            sw.WriteLine("module " + GetNamespace(config, root) + " {");
            sw.WriteLine();
        }
    }

    public void WriteNamespaceEnd(IJsonClassGeneratorConfig config, TextWriter sw, bool root)
    {
        if (GetNamespace(config, root) != null)
        {
            sw.WriteLine("}");
            sw.WriteLine();
        }
    }

}