﻿// Copyright © 2010 Xamasoft

using Newtonsoft.Json.Linq;

namespace Json2CSharpCodeGenerator.Lib;

public class JsonType
{

    private JsonType(IJsonClassGeneratorConfig generator) => this._Generator = generator;

    public JsonType(IJsonClassGeneratorConfig generator, JToken token)
        : this(generator)
    {

        Type = GetFirstTypeEnum(token);

        if (Type == JsonTypeEnum.Array)
        {
            JArray array = (JArray)token;
            InternalType = GetCommonType(generator, array.ToArray());
        }
    }

    internal static JsonType GetNull(IJsonClassGeneratorConfig generator) => new(generator, JsonTypeEnum.NullableSomething);

    private readonly IJsonClassGeneratorConfig _Generator;

    internal JsonType(IJsonClassGeneratorConfig generator, JsonTypeEnum type)
        : this(generator) => Type = type;

    public static JsonType GetCommonType(IJsonClassGeneratorConfig generator, JToken[] tokens)
    {

        if (tokens.Length == 0)
            return new JsonType(generator, JsonTypeEnum.NonConstrained);

        JsonType common = new JsonType(generator, tokens[0]).MaybeMakeNullable(generator);

        for (int i = 1; i < tokens.Length; i++)
        {
            JsonType current = new(generator, tokens[i]);

            try
            {
                common = common.GetCommonType(current);
            }
            catch (Exception)
            {
                throw;
            }
        }

        return common;

    }

    internal JsonType MaybeMakeNullable(IJsonClassGeneratorConfig generator)
    {
        if (!generator.AlwaysUseNullableValues)
            return this;
        return GetCommonType(GetNull(generator));
    }

    public JsonTypeEnum Type { get; private set; }
    public JsonType InternalType { get; private set; }
    public string AssignedName { get; private set; }
    public string OriginalName { get; private set; }
    public string NewAssignedName { get; private set; }

    public void AssignName(string name) => AssignedName = name;
    public void AssignOriginalName(string name) => OriginalName = name;
    public void AssignNewAssignedName(string name) => NewAssignedName = name;

    public bool MustCache
    {
        get
        {
            return Type switch
            {
                JsonTypeEnum.Array => true,
                JsonTypeEnum.Object => true,
                JsonTypeEnum.Anything => true,
                JsonTypeEnum.Dictionary => true,
                JsonTypeEnum.NonConstrained => true,
                _ => false,
            };
        }
    }

    public string GetReaderName()
    {
        if (Type is JsonTypeEnum.Anything or JsonTypeEnum.NullableSomething or JsonTypeEnum.NonConstrained)
        {
            return "ReadObject";
        }
        if (Type == JsonTypeEnum.Object)
        {
            return string.Format("ReadStronglyTypedObject<{0}>", AssignedName);
        }
        else if (Type == JsonTypeEnum.Array)
        {
            return string.Format("ReadArray<{0}>", InternalType.GetTypeName());
        }
        else
        {
            return string.Format("Read{0}", Enum.GetName(typeof(JsonTypeEnum), Type));
        }
    }

    public JsonType GetInnermostType()
    {
        if (Type != JsonTypeEnum.Array)
            throw new InvalidOperationException();
        if (InternalType.Type != JsonTypeEnum.Array)
            return InternalType;
        return InternalType.GetInnermostType();
    }

    public bool IsCollectionType() => Type is JsonTypeEnum.Array or JsonTypeEnum.Dictionary;

    public string GetTypeName() => _Generator.CodeWriter.GetTypeName(this, _Generator);

    public string GetJTokenType()
    {
        return Type switch
        {
            JsonTypeEnum.Boolean or JsonTypeEnum.Integer or JsonTypeEnum.Long or JsonTypeEnum.Float or JsonTypeEnum.Date or JsonTypeEnum.NullableBoolean or JsonTypeEnum.NullableInteger or JsonTypeEnum.NullableLong or JsonTypeEnum.NullableFloat or JsonTypeEnum.NullableDate or JsonTypeEnum.String => "JValue",
            JsonTypeEnum.Array => "JArray",
            JsonTypeEnum.Dictionary => "JObject",
            JsonTypeEnum.Object => "JObject",
            _ => "JToken",
        };
    }

    public JsonType GetCommonType(JsonType type2)
    {

        JsonTypeEnum commonType = GetCommonTypeEnum(Type, type2);

        if (commonType == JsonTypeEnum.Array)
        {
            if (type2.Type == JsonTypeEnum.NullableSomething)
                return this;
            if (Type == JsonTypeEnum.NullableSomething)
                return type2;

            JsonType commonInternalType;
            if (InternalType == null && type2.InternalType != null) // Handling the case Test_4 where the first array is an empty object
                commonInternalType = type2.InternalType;
            else
                commonInternalType = InternalType.GetCommonType(type2.InternalType).MaybeMakeNullable(_Generator);

            if (commonInternalType != InternalType)
                return new JsonType(_Generator, JsonTypeEnum.Array) { InternalType = commonInternalType };
        }

        //if (commonType == JsonTypeEnum.Dictionary)
        //{
        //    var commonInternalType = InternalType.GetCommonType(type2.InternalType);
        //    if (commonInternalType != InternalType) return new JsonType(JsonTypeEnum.Dictionary) { InternalType = commonInternalType };
        //}

        if (Type == commonType)
            return this;
        return new JsonType(_Generator, commonType).MaybeMakeNullable(_Generator);
    }

    private static bool IsNull(JsonTypeEnum type) => type == JsonTypeEnum.NullableSomething;

    private JsonTypeEnum GetCommonTypeEnum(JsonTypeEnum type1, JsonType type2json)
    {
        if (type2json == null)
            return type1;

        JsonTypeEnum type2 = type2json.Type;

        if (type1 == JsonTypeEnum.NonConstrained)
            return type2;
        if (type2 == JsonTypeEnum.NonConstrained)
            return type1;

        switch (type1)
        {
            case JsonTypeEnum.Boolean:
                if (IsNull(type2))
                    return JsonTypeEnum.NullableBoolean;
                if (type2 == JsonTypeEnum.Boolean)
                    return type1;
                break;
            case JsonTypeEnum.NullableBoolean:
                if (IsNull(type2))
                    return type1;
                if (type2 == JsonTypeEnum.Boolean)
                    return type1;
                break;
            case JsonTypeEnum.Integer:
                if (IsNull(type2))
                    return JsonTypeEnum.NullableInteger;
                if (type2 == JsonTypeEnum.Float)
                    return JsonTypeEnum.Float;
                if (type2 == JsonTypeEnum.Long)
                    return JsonTypeEnum.Long;
                if (type2 == JsonTypeEnum.Integer)
                    return type1;
                break;
            case JsonTypeEnum.NullableInteger:
                if (IsNull(type2))
                    return type1;
                if (type2 == JsonTypeEnum.Float)
                    return JsonTypeEnum.NullableFloat;
                if (type2 == JsonTypeEnum.Long)
                    return JsonTypeEnum.NullableLong;
                if (type2 == JsonTypeEnum.Integer)
                    return type1;
                break;
            case JsonTypeEnum.Float:
                if (IsNull(type2))
                    return JsonTypeEnum.NullableFloat;
                if (type2 == JsonTypeEnum.Float)
                    return type1;
                if (type2 == JsonTypeEnum.Integer)
                    return type1;
                if (type2 == JsonTypeEnum.Long)
                    return type1;
                break;
            case JsonTypeEnum.NullableFloat:
                if (IsNull(type2))
                    return type1;
                if (type2 == JsonTypeEnum.Float)
                    return type1;
                if (type2 == JsonTypeEnum.Integer)
                    return type1;
                if (type2 == JsonTypeEnum.Long)
                    return type1;
                break;
            case JsonTypeEnum.Long:
                if (IsNull(type2))
                    return JsonTypeEnum.NullableLong;
                if (type2 == JsonTypeEnum.Float)
                    return JsonTypeEnum.Float;
                if (type2 == JsonTypeEnum.Integer)
                    return type1;
                break;
            case JsonTypeEnum.NullableLong:
                if (IsNull(type2))
                    return type1;
                if (type2 == JsonTypeEnum.Float)
                    return JsonTypeEnum.NullableFloat;
                if (type2 == JsonTypeEnum.Integer)
                    return type1;
                if (type2 == JsonTypeEnum.Long)
                    return type1;
                break;
            case JsonTypeEnum.Date:
                if (IsNull(type2))
                    return JsonTypeEnum.NullableDate;
                if (type2 == JsonTypeEnum.Date)
                    return JsonTypeEnum.Date;
                break;
            case JsonTypeEnum.NullableDate:
                if (IsNull(type2))
                    return type1;
                if (type2 == JsonTypeEnum.Date)
                    return type1;
                break;
            case JsonTypeEnum.NullableSomething:
                if (IsNull(type2))
                    return type1;
                if (type2 == JsonTypeEnum.String)
                    return JsonTypeEnum.String;
                if (type2 == JsonTypeEnum.Integer)
                    return JsonTypeEnum.NullableInteger;
                if (type2 == JsonTypeEnum.Float)
                    return JsonTypeEnum.NullableFloat;
                if (type2 == JsonTypeEnum.Long)
                    return JsonTypeEnum.NullableLong;
                if (type2 == JsonTypeEnum.Boolean)
                    return JsonTypeEnum.NullableBoolean;
                if (type2 == JsonTypeEnum.Date)
                    return JsonTypeEnum.NullableDate;
                if (type2 == JsonTypeEnum.Array)
                    return JsonTypeEnum.Array;
                if (type2 == JsonTypeEnum.Object)
                    return JsonTypeEnum.Object;
                break;
            case JsonTypeEnum.Object:
                if (IsNull(type2))
                    return type1;
                if (type2 == JsonTypeEnum.Object)
                    return type1;
                if (type2 == JsonTypeEnum.Dictionary)
                    throw new ArgumentException();
                break;
            case JsonTypeEnum.Dictionary:
                throw new ArgumentException();
            //if (IsNull(type2)) return type1;
            //if (type2 == JsonTypeEnum.Object) return type1;
            //if (type2 == JsonTypeEnum.Dictionary) return type1;
            //  break;
            case JsonTypeEnum.Array:
                if (IsNull(type2))
                    return type1;
                if (type2 == JsonTypeEnum.Array)
                    return type1;
                break;
            case JsonTypeEnum.String:
                if (IsNull(type2))
                    return type1;
                if (type2 == JsonTypeEnum.String)
                    return type1;
                break;
        }

        return JsonTypeEnum.Anything;

    }

    private static bool IsNull(JTokenType type) => type is JTokenType.Null or JTokenType.Undefined;

    private static JsonTypeEnum GetFirstTypeEnum(JToken token)
    {
        JTokenType type = token.Type;
        if (type == JTokenType.Integer)
        {
            try
            {
                if ((long)((JValue)token).Value < int.MaxValue)
                    return JsonTypeEnum.Integer;
                else
                    return JsonTypeEnum.Long;
            }
            catch (Exception)
            {
                // Must be a BigInteger Number, either way make it as string
                return JsonTypeEnum.String;
            }

        }
        return type switch
        {
            JTokenType.Array => JsonTypeEnum.Array,
            JTokenType.Boolean => JsonTypeEnum.Boolean,
            JTokenType.Float => JsonTypeEnum.Float,
            JTokenType.Null => JsonTypeEnum.NullableSomething,
            JTokenType.Undefined => JsonTypeEnum.NullableSomething,
            JTokenType.String => JsonTypeEnum.String,
            JTokenType.Object => JsonTypeEnum.Object,
            JTokenType.Date => JsonTypeEnum.Date,
            _ => JsonTypeEnum.Anything,
        };
    }

    public List<FieldInfo> Fields { get; internal set; }
    public bool IsRoot { get; internal set; }
}